#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2010, 2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path

def get_word(word):
    return get_bytes(word % 256, word // 256)

def get_bytes(*bytes):
    return ''.join([chr(b) for b in bytes])

def get_parity(block):
    parity = 0
    for c in block[2:]:
        parity ^= ord(c)
    return get_bytes(parity)

def get_basic_loader(title):
    header = get_word(19)          # Length of header block
    header += get_bytes(0)         # Header block marker
    header += get_bytes(0)         # BASIC program follows
    header += title[:10].ljust(10) # Title padded with spaces up to 10 chars
    header += get_word(27)         # Length of entire data block
    header += get_word(10)         # LINE number to RUN after LOADing
    header += get_word(27)         # Length of BASIC program only
    header += get_parity(header)

    data = get_word(29)            # Length of data block
    data += get_bytes(255)         # Data block marker
    data += get_bytes(0, 10)       # Line 10
    data += get_word(5)            # Length of line 10
    data += get_bytes(239)         # LOAD
    data += get_bytes(34, 34)      # ""
    data += get_bytes(175)         # CODE
    data += get_bytes(13)          # ENTER
    data += get_bytes(0, 20)       # Line 20
    data += get_word(14)           # Length of line 20
    data += get_bytes(249)         # RANDOMIZE
    data += get_bytes(192)         # USR
    data += "23296"                # 23296
    data += get_bytes(14)          # Floating-point number marker
    data += get_bytes(0, 0)        # 23296 in
    data += get_word(23296)        # floating-point
    data += get_bytes(0)           # form
    data += get_bytes(13)          # ENTER
    data += get_parity(data)

    return header + data

def get_data_loader(title, org, length, start):
    header = get_word(19)          # Length of header block
    header += get_bytes(0)         # Header block marker
    header += get_bytes(3)         # CODE block follows
    header += title[:10].ljust(10) # Title padded with spaces up to 10 chars
    header += get_word(18)         # Length of data in data block
    header += get_word(23296)      # Start address
    header += get_word(0)          # Unused
    header += get_parity(header)

    data = get_word(20)                        # Length of data block
    data += get_bytes(255)                     # Data block marker
    data += get_bytes(221, 33) + get_word(org) # LD IX,ORG
    data += get_bytes(17) + get_word(length)   # LD DE,LENGTH
    data += get_bytes(55)                      # SCF
    data += get_bytes(159)                     # SBC A,A
    data += get_bytes(221, 249)                # LD SP,IX
    data += get_bytes(1) + get_word(start)     # LD BC,START
    data += get_bytes(197)                     # PUSH BC
    data += get_bytes(195, 86, 5)              # JP 1366
    data += get_parity(data)

    return header + data

def parse_options(options, binfile, binlength):
    org = 65536 - binlength
    start = None
    suffix = '.bin'
    if binfile.endswith(suffix):
        prefix = binfile[:-len(suffix)]
    else:
        prefix = binfile
    tapfile = prefix + ".tap"
    i = 0
    while i < len(options):
        option = options[i]
        value = options[i + 1]
        if option == '-o':
            org = int(value)
        elif option == '-s':
            start = int(value)
        elif option == '-t':
            tapfile = value
        i += 2
    return org, start or org, tapfile

#
# Begin
#
if len(sys.argv) < 2:
    sys.stderr.write("""Usage: %s [options] FILE.bin

  Convert a binary snapshot file into a TAP file.

Options:
  -o ORG      Set the origin (default: 65536 - length of FILE.bin)
  -s START    Set the start address to JP to (default: ORG)
  -t TAPFILE  Set the TAP filename (default: FILE.tap)
""" % os.path.basename(sys.argv[0]))
    sys.exit(1)

binfile = sys.argv[-1]
f = open(binfile, 'rb')
ram = f.read()
f.close()

length = len(ram)
org, start, tapfile = parse_options(sys.argv[1:-1], binfile, length)
data = get_word(length + 2) # Length of data block
data += get_bytes(255)      # Data block marker
data += ram                 # Data
data += get_parity(data)

tap_data = get_basic_loader(binfile)
tap_data += get_data_loader(binfile, org, length, start)
tap_data += data

f = open(tapfile, 'w')
f.write(tap_data)
f.close()
