# -*- coding: utf-8 -*-

# Copyright 2008-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import cgi
import math
from microsphere import MicrosphereGame, MicrosphereHtmlWriter, Udg
from skoollib import MAIN_CODE_ID
from skoolhtml import FileInfo, Table, Cell, join

class BackToSkool(MicrosphereGame):
    def __init__(self, skool_parser, ref_parser, code_id=MAIN_CODE_ID):
        MicrosphereGame.__init__(self, skool_parser, ref_parser, code_id)
        self.min_lesson = 37
        self.max_lesson = 59
        self.min_character = 183
        self.max_character = 209
        self.font_address = 55040
        self.char_buf_descs = self._get_char_buf_descs()
        self.keypress_routines = self.get_data_dictionary('KeypressRoutines')
        self.as_descs = self.get_data_dictionary('AnimatoryStates')
        self.characters = self.get_data_dictionary('Characters')
        self.lesson_descs = self.get_data_dictionary('Lessons')
        self._calculate_tap_index(59392, 88)

    def get_skool_udg(self, y, x, show_chars=False):
        q_addr = x + 46336
        q = self.snapshot[q_addr]
        h = y + 160
        hl = 180 + q // 2 + 256 * h
        i = 0
        if q & 1:
            p = self.snapshot[hl] & 15
        else:
            p = (self.snapshot[hl] & 240) // 16
        if p == 0:
            i = q & 3 or 2
            if q & 1:
                p = self.snapshot[hl + 256] & 15
            else:
                p = (self.snapshot[hl + 256] & 240) // 16
        attr = i | (8 * p)
        ref_addr = q + 256 * h
        ref = self.snapshot[ref_addr]
        a = 136 // int(math.pow(2, q & 3)) & self.snapshot[144 + q // 4 + 256 * h]
        udg_page = (8 if a & 15 else 0) + (144 if a & 240 else 128)
        udg_addr = ref + 256 * udg_page
        udg_bytes = self.snapshot[udg_addr:udg_addr + 2048:256]
        skool_udg = Udg(attr, udg_bytes, q_addr=q_addr, ref_addr=ref_addr, ref=ref, udg_page=udg_page, x=x, y=y)
        if show_chars:
            self._superimpose_sprite_udgs(skool_udg, x, y)
        return skool_udg

    def get_sprite_udg(self, state, attr, row, col, udg_page=None):
        as_norm = state | 128
        udg_page = 183 if as_norm < 208 else 199
        col = 2 - col if state & 128 else col
        ref_page = 215 + col * 4 + row
        udg = self._get_sprite_udg(as_norm, attr, ref_page, udg_page)
        if state & 128:
            self._flip_udg(udg)
        return udg

    def _alter_skool_udgs(self, address):
        coords = []
        while True:
            y = self.snapshot[address]
            if y == 255:
                break
            x = self.snapshot[address + 1]
            coords.append((x, y))
            udg_lsb = self.snapshot[address + 2]
            attr = self.snapshot[address + 3]
            d = y + 160
            col_ref = self.snapshot[x + 46336]
            self.snapshot[col_ref + 256 * d] = udg_lsb
            z = int(math.pow(2, col_ref & 3))
            e = 128 + (col_ref + 104) // 2
            a = self.snapshot[e + 256 * d]
            if col_ref & 1:
                a = (a & 240) | (attr & 15)
            else:
                a = (16 * (attr & 15)) | (a & 15)
            self.snapshot[e + 256 * d] = a
            a = attr & 192
            if a & 64:
                a = a - 56
            e = 144 + col_ref // 4
            self.snapshot[e + 256 * d] = self.snapshot[e + 256 * d] & 255 - 136 // z | a // z
            address += 4
        return coords

    def _get_skool_udgs_by_coords(self, coords):
        left_x, top_y, width, height = self._get_dimensions(coords)
        udg_array = []
        for y in range(top_y, top_y + height):
            udg_array.append([None] * width)
        for x, y in coords:
            udg_array[y - top_y][x - left_x] = self.get_skool_udg(y, x)
        return udg_array

    def _get_dimensions(self, coords, padding=0):
        x_range = [c[0] for c in coords]
        y_range = [c[1] for c in coords]
        left_x = min(x_range)
        top_y = min(y_range)
        width = max(x_range) - left_x + 1
        height = max(y_range) - top_y + 1
        left_x -= padding
        top_y -= padding
        width += 2 * padding if left_x + width < 192 else padding
        height += 2 * padding if top_y + height < 20 else padding
        return left_x, top_y, width, height

    def get_udg(self, address, attr):
        return Udg(attr, self.snapshot[address:address + 8])

    def get_mutable_udg_array(self, address, padding):
        self.push_snapshot()
        coords = self._alter_skool_udgs(address)
        if padding >= 0:
            udg_array = self.get_skool_udgs(*self._get_dimensions(coords, padding))
        else:
            udg_array = self._get_skool_udgs_by_coords(coords)
        self.pop_snapshot()
        return udg_array

    def get_initial_states(self):
        return [self.snapshot[32 + 256 * c:35 + 256 * c] for c in range(183, 211)]

class BackToSkoolWriter(MicrosphereHtmlWriter):
    def write_mutable_tile_tables(self, cwd):
        lines = []

        doors = []
        doors.append((56637, True, 'Left study door'))
        doors.append((57088, True, 'Right study door'))
        doors.append((56576, True, 'Science Lab storeroom door'))
        doors.append((57149, False, "Boys' skool door"))
        doors.append((63744, False, 'Skool gate'))
        doors.append((56064, True, 'Drinks cabinet door'))

        windows = []
        windows.append((55040, True, 'Top floor window'))
        windows.append((55552, True, 'Middle floor window'))

        bike = [(57600, False, 'Tree without/with bike')]

        sections = (('Doors', doors), ('Windows (open)', windows), ('Tree (without bike)', bike))
        for title, section in sections:
            lines.append('<h3>%s</h3>\n<table class="mgrid">\n<tr>\n' % title)
            for a, initial_state, desc in section:
                addr = a + 256 if initial_state else a
                mutable_udgs = self.skool.get_mutable_udg_array(addr, -1)
                lines.append('<td>\n')
                lines += self.write_playarea_tile_table(cwd, mutable_udgs, 'm')
                lines.append('</td>\n')
            lines.append('</tr>\n</table>\n')

        cups = []
        cups.append((56694, "Leftmost cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((56950, "Middle cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((57206, "Rightmost cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((57462, "Cup on the shelf in the girls' skool (water/sherry)"))
        lines.append('<h3>Cups</h3>\n<table>\n')
        for a, desc, in cups:
            lines.append('<tr>\n<td>\n<table class="mgrid">\n<tr>\n')
            for address in (a, a + 5):
                cup_udg_array = self.skool.get_mutable_udg_array(address, -1)
                lines.append('<td>\n')
                lines += self.write_playarea_tile_table(cwd, cup_udg_array, 'c%i_' % address)
                lines.append('</td>\n')
            lines.append('</tr>\n</table>\n</td><td>%s</td></tr>\n' % desc)
        lines.append('</table>\n')

        return lines

    def keypress_table(self, cwd):
        keypress_routines = self.skool.keypress_routines
        snapshot = self.skool.snapshot

        table = Table('keytable', 'Index', 'Key', 'Address', 'Offset', 'Lookup', 'Routine', 'Purpose')
        for index in range(80):
            key = '&#169;' if index == 79 else cgi.escape(chr(index + 48))
            address = index + 58624
            offset = snapshot[address]
            lookup = routine_link = purpose = ''
            if offset:
                lookup = offset + 58624
                routine = snapshot[lookup] + 256 * snapshot[lookup + 1]
                purpose = keypress_routines[routine]
                routine_link = self.expand('#R%i' % routine, cwd)
            row = []
            row.append(Cell(index, "keyIndex"))
            row.append(Cell(key, "key"))
            row.append(Cell(address, "keyAddress"))
            row.append(Cell(offset, "data"))
            row.append(Cell(lookup, "keyAddress"))
            row.append(Cell(routine_link, "keyAddress"))
            row.append(Cell(purpose, "purpose"))
            table.add_row(row)

        return table.html()

    def get_playarea_tile_text(self, tile):
        text = ['<i>Q_addr</i>=(x,181)=%i, <i>Q</i>=%i' % (tile.q_addr, self.skool.snapshot[tile.q_addr])]
        text.append('<i>ref_addr</i>=(<i>Q</i>,y+160)=%i' % tile.ref_addr)
        text.append('<i>ref</i>=%i, <i>udg_page</i>=%i' % (tile.ref, tile.udg_page))
        text.append('<i>udg_addr</i>=(<i>ref</i>,<i>udg_page</i>)=%i' % (tile.ref + 256 * tile.udg_page))
        text.append('<i>attr</i>=%i' % tile.attr)
        return text

    def mutable(self, cwd, address, padding=0):
        fname = 'mutable%i_%i' % (address, padding) if padding else 'mutable%i' % address
        mutable_fname = join('images', 'mutables', '%s.png' % fname)
        if self.need_image(mutable_fname):
            mutable_udgs = self.skool.get_mutable_udg_array(address, padding)
            self.write_image(mutable_udgs, mutable_fname, 2)
        img_path = FileInfo.relpath(cwd, mutable_fname)
        return '<img alt="%s" src="%s" />' % (fname, img_path)
